/* NiuTrans.Tensor - an open-source tensor library
 * Copyright (C) 2018, Natural Language Processing Lab, Northestern University.
 * All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/*
* $Created by: HU Chi (huchinlp@foxmail.com) 2019-04-05
*/

#include "DataSet.h"
#include "StringUtil.h"

#include <string>
#include <vector>
#include <fstream>
#include <algorithm>


using namespace nts;

const int PAD = 0;

/*
load data from the file to the buffer
the data format: 
each line: field1 ||| field2 ||| ... ||| fieldn
i.e. fields are separated by `|||` and tokens are separated by space
this will sort the data in ascending order if `sortBuffer` is True
*/
void DataSet::LoadDataToBuffer()
{
    string line;
    buffer.clear();
    bufferUsed = 0;
    const string tokenDelimiter = " ";
    const string fieldsDelimiter = "|||";

    int counter = bufferSize;
    while (getline(*fp, line) && counter) {
        auto fields = Split<string>(line, fieldsDelimiter);
        for (const auto& field : fields) {
            auto elements = Split<int>(field, tokenDelimiter);
            buffer.emplace_back(elements);
        }
        counter--;
    }
    if (fp->eof()) {
        fp->seekg(fp->beg);
        // LOG("start a new epoch");
	}
    //if (sortBuffer) {

    //    auto myCompare = [](const auto& a, const auto& b) {
    //        return a.first.size() < b.first.size();
    //    };
    //    sort(buffer.begin(), buffer.end(), myCompare);
    //}
}

/*
select a field and generate a mini-batch by indices
>>> t - a tensor to store the batch
>>> indices - the indices of data
>>> offset - the number of indices
>>> field - indicates which field is selected
this will combine selected elements into a padded batch
*/
void DataSet::LoadBatch(XTensor& t, const int* indices, size_t batchSize, size_t field)
{
    if (bufferUsed == bufferSize || bufferUsed == 0) {
        LoadDataToBuffer();
	}

    /* get the maximum length in a mini-batch */
    size_t maxLength = 0;
    for (size_t i = 0; i < batchSize; ++i) {
        int idx = *(indices + i);
        maxLength = max(maxLength, buffer[idx * fieldNum + field].size());
    }

    int* data = new int[maxLength * batchSize];
    memset(data, 0, maxLength * batchSize);

    size_t cur = 0;
    for (size_t i = 0; i < batchSize; ++i) {
        size_t next = cur + maxLength;
        int idx = *(indices + i);
        for (int v : buffer[idx * fieldNum + field]) {
            data[cur++] = v;
		}
        /* pad zeros */
        while (cur < next) {
            data[cur++] = 0;
        }
    }
    InitTensor2D(&t, batchSize, maxLength, X_INT);
    t.SetData(data, maxLength * batchSize);
    bufferUsed += batchSize;
    delete[] data;
}

/*
the constructor of DataSet
>>> fname - path of the data file
>>> paraFieldNum - the number of different fields
>>> paraBufferSize - size of each field in the buffer
the real size of buffer is `bufferSize * fieldNum`
*/
DataSet::DataSet(const char* fname, size_t paraFieldNum, size_t paraBufferSize)
{
    fp = new ifstream(fname);
    fieldNum = paraFieldNum;
    bufferSize = paraBufferSize;
    bufferUsed = 0;
    buffer.reserve(bufferSize * fieldNum);
    CheckNTErrors(fp, "unable to open the file: %s", fname);
}
