/* NiuTrans.Tensor - an open-source tensor library
 * Copyright (C) 2018, Natural Language Processing Lab, Northestern University.
 * All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/*
* $Created by: HU Chi (huchinlp@foxmail.com) 2019-03-18
*/

#include "Vocabulary.h"

#include "StringUtil.h"
#include <algorithm>
#include <fstream>
#include <iterator>
#include <map>
#include <string>
#include <vector>
using namespace nts;

/* initialize the mappings with some special tokens */
void nts::Vocabulary::Init()
{
    const std::map<std::string, int> init{
        { "<UNK>", 0 },
        { "<PAD>", 1 },
        { "<SOS>", 2 },
        { "<EOS>", 3 },
    };
    for (const auto& v : init) {
        token2ID[v.first] = v.second;
        id2Token[v.second] = v.first;
    }
}

/*
load vocabulary from disk
format(each line): token<\n>
*/
nts::Vocabulary::Vocabulary(const char* vocabPath)
{
    Init();
    std::ifstream file(vocabPath);
    string line;
    int index = int(token2ID.size());
    while (getline(file, line)) {
        size_t pos = line.find("\n");
        string token = line.substr(0, pos);
        token2ID[token] = index;
        id2Token[index++] = token;
    }
    currentSize = int(token2ID.size());
}

/* 
build vocabulary from the source file
>> srcPath - the path of source file
>> minFreq - the user specified minimum count for a single token
>> maxSize - the user specified maxmum size of a vocabulary
this will sort tokens by frequency and map the most uncommon tokens to <UNK>
*/
nts::Vocabulary::Vocabulary(const char* srcPath, int minFreq, int maxSize)
{
    Init();
    using Dict = vector<pair<string, int>>;

    /* count word and their frequency  */
    std::ifstream file(srcPath);
    map<string, int> dict;
    string token;
    while (file >> token) {
        dict[token] += 1;
    }

    /* sort tokens by frequency */
    Dict dictFlatten;
    copy(dict.begin(), dict.end(), back_inserter<Dict>(dictFlatten));
    auto myCompare = [](const auto& a, const auto& b) {
        return a.second < b.second;
    };
    sort(dictFlatten.begin(), dictFlatten.end(), myCompare);

    /* skip token whose frequency is out of top-maxSize or less than minFreq */
    size_t index = token2ID.size();
    size_t validSize = maxSize > dict.size() ? maxSize : dict.size();
    for (const auto& tokenID : dictFlatten) {
        if (validSize && tokenID.second > minFreq && 
			token2ID.find(tokenID.first) == token2ID.end()) {
            id2Token[int(index)] = tokenID.first;
            token2ID[tokenID.first] = int(index);
            index++;
            validSize--;
        }
    }
    currentSize = int(token2ID.size());
}

/* insert a token to the vocabulary */
void nts::Vocabulary::Insert(string token)
{
    if (token2ID.find(token) != token2ID.end()) {
        token2ID[token] = currentSize;
        id2Token[currentSize++] = token;
    }
}

/* save the vocabulary to a file */
void nts::Vocabulary::Dump(const char* vocabPath)
{
    std::ofstream file(vocabPath);

    for (const auto& tokenID : token2ID)
        file << tokenID.first << "\t" << tokenID.second << "\n";
}

/* 
maps tokens to integers 
>>> tokens - a list of tokens
<<< indices - a list of corresponding indices
notices that this will map OOV to UNK
*/
vector<int> nts::Vocabulary::Token2ID(vector<string> tokens)
{
    vector<int> indices;
    indices.reserve(tokens.size());
    for (auto str : tokens) {
        if (token2ID.find(str) == token2ID.end())
            indices.emplace_back(token2ID["<UNK>"]);
        else
            indices.emplace_back(token2ID[str]);
    }
    return indices;
}

/* 
maps integers to tokens 
>>> indices - a list of indices
<<< tokens - a list of corresponding tokens
notices that this will throw a error if the id is not found
*/
vector<string> nts::Vocabulary::ID2Token(vector<int> ids)
{
    vector<string> tokens;
    tokens.reserve(ids.size());
    for (auto id : ids) {
        CheckNTErrors(id2Token.find(id) != id2Token.end(), "id not found!");
        tokens.emplace_back(id2Token[id]);
    }
    return tokens;
}
