/* NiuTrans.Tensor - an open-source tensor library
 * Copyright (C) 2018, Natural Language Processing Lab, Northestern University.
 * All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/*
* $Created by: HU Chi (huchinlp@foxmail.com) 2019-03-18
*/

#include "DataReader.h"

#include "StringUtil.h"
#include <algorithm>
#include <fstream>
#include <iterator>
#include <map>
#include <string>
#include <vector>
using namespace nts;


/* initialize the mappings with some special tokens */
void nts::Vocabulary::Init()
{
    const std::map<std::string, int> init{
        { "<UNK>", 0 },
        { "<PAD>", 1 },
        { "<SOS>", 2 },
        { "<EOS>", 3 },
    };
    for (const auto& v : init) {
        token2ID[v.first] = v.second;
        id2Token[v.second] = v.first;
    }
}

/*
load vocabulary from disk
format(each line): token<\n>
notice that there are some pre-defined special tokens
*/
nts::Vocabulary::Vocabulary(const char* vocabPath)
{
    Init();
    std::ifstream file(vocabPath);
    string line;
    int index = int(token2ID.size());
    while (getline(file, line)) {
        size_t pos = line.find("\n");
        string token = line.substr(0, pos);
        token2ID[token] = index;
        id2Token[index++] = token;
    }
    currentSize = int(token2ID.size());
}

/* 
build vocabulary from the source file
>> srcPath - the path of source file
>> minFreq - the user specified minimum count for a single token
>> maxSize - the user specified maxmum size of a vocabulary
we sort tokens by their frequency, the most uncommon tokens will be mapped to <UNK>
if the amount of different tokens is larger than paraMaxSize.
*/
nts::Vocabulary::Vocabulary(const char* srcPath, int minFreq, int maxSize)
{
    Init();

    /* count word and their frequency  */
    std::ifstream file(srcPath);
    map<string, int> dict;
    string token;
    while (file >> token) {
        dict[token] += 1;
    }

    /* sort word by their frequency */
    vector<pair<string, int>> dictFlatten;
    copy(dict.begin(), dict.end(), back_inserter<vector<pair<string, int>>>(dictFlatten));
    auto myCompare = [](const auto& a, const auto& b) {
        return a.second < b.second;
    };
    sort(dictFlatten.begin(), dictFlatten.end(), myCompare);

    /* skip token whose frequency is out of top-maxSize or less than minFreq */
    size_t index = token2ID.size();
    size_t validSize = maxSize > dict.size() ? maxSize : dict.size();
    for (const auto& tokenID : dictFlatten) {
        if (validSize && tokenID.second > minFreq && token2ID.find(tokenID.first) == token2ID.end()) {
            id2Token[int(index)] = tokenID.first;
            token2ID[tokenID.first] = int(index);
            index++;
            validSize--;
        }
    }
    currentSize = int(token2ID.size());
}

/* insert a token to the vocabulary */
void nts::Vocabulary::Insert(string token)
{
    if (token2ID.find(token) != token2ID.end()) {
        token2ID[token] = currentSize;
        id2Token[currentSize++] = token;
    }
}

/* save the vocabulary to a file */
void nts::Vocabulary::Dump(const char* vocabPath)
{
    std::ofstream file(vocabPath);

    for (const auto& tokenID : token2ID)
        file << tokenID.first << "\t" << tokenID.second << "\n";
}

/* 
maps tokens to integers 
>>> tokens - a list of tokens
<<< indices - a list of corresponding indices
notices that this will map OOV to UNK
*/
vector<int> nts::Vocabulary::Token2ID(vector<string> tokens)
{
    vector<int> indices;
    indices.reserve(tokens.size());
    for (auto str : tokens) {
        if (token2ID.find(str) == token2ID.end())
            indices.emplace_back(token2ID["<UNK>"]);
        else
			indices.emplace_back(token2ID[str]);
	}
    return indices;
}

/* 
maps integers to tokens 
>>> indices - a list of indices
<<< tokens - a list of corresponding tokens
notices that this will throw a error if the id is not found
*/
vector<string> nts::Vocabulary::ID2Token(vector<int> ids)
{
    vector<string> tokens;
    tokens.reserve(ids.size());
    for (auto id : ids) {
        CheckNTErrors(id2Token.find(id) != id2Token.end(), "id not found!");
        tokens.emplace_back(id2Token[id]);
	}
    return tokens;
}

/* 
shuffle lines of the file 
>> srcFile - the source file to shuffle
this will create a file "<srcFile>.random"
*/
void nts::Shuffle(const char* srcFile)
{
    string command = "shuf ";
    string tgtFile = srcFile;
    tgtFile += ".random";
    command += srcFile;
    command += " > ";
    command += tgtFile;
#ifndef WIN32
    system(command.c_str());
#else
    ShowNTErrors("Cannot shuffle the file on WINDOWS systems!");
#endif
}

/* 
load data to the buffer
data format(a single line): "s1 s2 ... sn|||t1 t2 ... tm" 
where s is the source token and t is the target token
this will transfer raw tokens to indices according to the vocabulary
>>> bufferSize - the number of sentences in the buffer
*/
void nts::DataSet::LoadDataToBuf()
{
    dataBuf.clear();
    string line;
    const string srcTgtDelimiter = "|||";
    const string sentDelimiter = " ";
    int counter = bufferSize;

    while (getline(*fp, line) && counter) {
        auto subStrings = Split<string>(line, srcTgtDelimiter);
        auto src = Split<string>(subStrings[0], sentDelimiter);
        auto tgt = Split<string>(subStrings[1], sentDelimiter);
        auto srcIndices = srcVocab->Token2ID(src);
        auto tgtIndices = tgtVocab->Token2ID(tgt);
        dataBuf.emplace_back(srcIndices, tgtIndices);
        counter--;
    }
}

/*
splits the data buffer to buckets
elements of the buffer are grouped together by length and batched
some samples may never be selected due to their unsuitable length
>>> batchSize - the maximum number of tokens
<<< buckets - a list of tokens indices in a mini-batch
*/
void nts::DataManager::BucketSeqByLen(int batchSize)
{
    dataSet->LoadDataToBuf();
    auto myCompare = [](const auto& a, const auto& b) {
        return a.first.size() < b.first.size();
    };
    sort(dataSet->dataBuf.begin(), dataSet->dataBuf.end(), myCompare);

    int realBatchSize = 0;
    for (int i = 0; i < dataSet->dataBuf.size(); ++i) {

        /* the first allocation */
        if (!buckets.size()) {
            buckets.emplace_back(1);
            buckets.back().pop_back();
        }

        /* append the current sentence to bucket  */
        size_t seqLen = dataSet->dataBuf[i].first.size();
        size_t seqNum = buckets.back().size();
        if ((seqNum + 1) * seqLen <= batchSize && 
			seqLen <= maxSeqLen && seqLen >= minSeqLen) {
            buckets.back().push_back(i);
        }

        /* allocate a new bucket */
        else {
            buckets.emplace_back(1);
            buckets.back().pop_back();
        }
    }

    /* drop the last bucket if its size is too small */
    if (buckets.back().size() * dataSet->dataBuf.back().first.size() < batchSize && buckets.size()>1)
        buckets.pop_back();
}

/*
load a batch of data
>> batchEnc - the batch of the input sequences
>> paddingEnc - padding of the input sequences
>> batchDec - the batch of the output sequences
>> paddingDec - padding of the output sequences
>> devID - device id
>> mem - memory pool
*/
void nts::DataManager::LoadBatch(XTensor* batchEnc, XTensor* paddingEnc, 
								 XTensor* batchDec, XTensor* paddingDec, 
								 XTensor* label, int devID, XMem* mem)
{
    if (bucketIter == buckets.size() - 1 || buckets.size() == 0) {
        BucketSeqByLen(batchSize);
        bucketIter = 0;
    }
    auto bucket = buckets[bucketIter];
    bucketIter++;
    size_t sentenceCount = bucket.size();
    size_t maxEnc = dataSet->dataBuf[bucket.back()].first.size();
    size_t maxDec = dataSet->dataBuf[bucket.back()].second.size();

    InitTensor2D(batchEnc, int(sentenceCount), int(maxEnc), X_INT, devID, mem);
    InitTensor2D(paddingEnc, int(sentenceCount), int(maxEnc), X_FLOAT, devID, mem);
    InitTensor2D(batchDec, int(sentenceCount), int(maxDec), X_INT, devID, mem);
    InitTensor2D(paddingDec, int(sentenceCount), int(maxDec), X_FLOAT, devID, mem);
    InitTensor2D(label, int(sentenceCount), int(maxDec), X_INT, devID, mem);

    int* labelValues = new int[batchDec->unitNum];
    int* batchEncValues = new int[batchEnc->unitNum];
    int* batchDecValues = new int[batchDec->unitNum];
    float* paddingEncOffsets = new float[batchEnc->unitNum];
    float* paddingDecOffsets = new float[batchDec->unitNum];

    memset(batchEncValues, 0, sizeof(int) * batchEnc->unitNum);
    memset(batchDecValues, 0, sizeof(int) * batchDec->unitNum);
    memset(paddingEncOffsets, 0, sizeof(float) * batchEnc->unitNum);
    memset(paddingDecOffsets, 0, sizeof(float) * batchDec->unitNum);
    memset(labelValues, 0, sizeof(float) * batchDec->unitNum);

    size_t srcIter = 0;
    size_t tgtIter = 0;

    for (const auto& index : bucket) {
        size_t srcOffset = srcIter + dataSet->dataBuf[bucket.back()].first.size();
        size_t tgtOffset = tgtIter + dataSet->dataBuf[bucket.back()].second.size();

        for (const auto& src : dataSet->dataBuf[index].first)
            batchEncValues[srcIter++] = src;

        for (size_t i = 0; i < dataSet->dataBuf[index].second.size(); ++i) {
            if (i != 0)
                labelValues[tgtIter - 1] = dataSet->dataBuf[index].second[i];
            batchDecValues[tgtIter++] = dataSet->dataBuf[index].second[i];
        }

        while (srcIter < srcOffset)
            paddingEncOffsets[srcIter++] = 1.0F;
        while (tgtIter < tgtOffset)
            paddingDecOffsets[tgtIter++] = 1.0F;
    }

    batchEnc->SetData(batchEncValues, batchEnc->unitNum);
    batchDec->SetData(batchDecValues, batchDec->unitNum);
    paddingEnc->SetData(paddingEncOffsets, batchEnc->unitNum);
    paddingDec->SetData(paddingDecOffsets, batchDec->unitNum);
    label->SetData(labelValues, label->unitNum);

    delete[] batchEncValues;
    delete[] batchDecValues;
    delete[] labelValues;
    delete[] paddingEncOffsets;
    delete[] paddingDecOffsets;
}
