/* NiuTrans.Tensor - an open-source tensor library
 * Copyright (C) 2018, Natural Language Processing Lab, Northestern University.
 * All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/*
* $Created by: HU Chi (huchinlp@foxmail.com) 2019-03-18
*/

#ifndef _DATA_READER_H_
#define _DATA_READER_H_

#include "../XGlobal.h"
#include "../XTensor.h"

#include <cstdio>
#include <fstream>
#include <unordered_map>

#define LOG(...)                      \
    do {                              \
        fprintf(stderr, "[INFO] ");   \
        fprintf(stderr, __VA_ARGS__); \
        fprintf(stderr, "\n");        \
        fflush(stdout);               \
    } while (0)

#define ERR(...)                      \
    do {                              \
        fprintf(stderr, "[ERROR] ");  \
        fprintf(stderr, __VA_ARGS__); \
        fprintf(stderr, "\n");        \
        fflush(stdout);               \
    } while (0)

using namespace std;

namespace nts { // namespace nts(NiuTrans.Tensor)

/* 
 * A Vocabulary maps strings to integers, allowing for strings to be mapped to an out-of-vocabulary token.
 * Vocabularies are fit to a particular dataset, which we use to decide which tokens are in-vocabulary.
 * For convinience, a vocabulary can be built from a source file or be loaded from a vocabulary file.
 */
class Vocabulary {

private:
    /* current size of the vocabulary */
    int currentSize = 0;

    /* initialize mappings with some special tokens */
    void Init();

public:
    /* a dict maps tokens to indices */
    unordered_map<string, int> token2ID;

    /* a dict maps indices to tokens */
    unordered_map<int, string> id2Token;

    Vocabulary() = delete;

    Vocabulary(Vocabulary&&) = delete;

    /* load the vocabulary from a file  */
    explicit Vocabulary(const char* vocabPath);

    /* built the vocabulary from the source file */
    explicit Vocabulary(const char* srcPath, int minFreq, int maxSize);

    /* append token to the vocabulary */
    void Insert(string token);

    /* save the vocabulary to a file */
    void Dump(const char* vocabPath);

    /* maps tokens to integers */
    vector<int> Token2ID(vector<string> tokens);

    /* maps integers to tokens */
    vector<string> ID2Token(vector<int> ids);
};

/* 
 * A DataSet maintains a sequences buffer.
 * An instance is tied to a specified file and buffer size.
 */
class DataSet {

public:
    /* number of buffered sequences */
    int bufferSize;

    /* the pointer to file stream */
    ifstream* fp;

    /* source-side vocabulary */
    Vocabulary* srcVocab;

    /* target-side vocabulary */
    Vocabulary* tgtVocab;

    /* sequences buffer */
    vector<pair<vector<int>, vector<int>>> dataBuf;

    DataSet(const DataSet&) = delete;

    DataSet(DataSet&&) = delete;

    /* constructor */
    explicit DataSet(const char* fname, const char* srcVocabPath, const char* tgtVocabPath, int paraBufferSize, bool isShuffled)
    {
        bufferSize = paraBufferSize;

        srcVocab = new Vocabulary(srcVocabPath);
        tgtVocab = new Vocabulary(tgtVocabPath);

        string trainFN = fname;
#ifndef WIN32
        if (isShuffled) {
            Shuffle(trainFN);
            trainFN += ".random";
        }
#endif
        fp = new ifstream();
        fp->open(trainFN);
        CheckNTErrors(fp, "unable to open %s", fname);
    }

    /* destructor */
    ~DataSet()
    {
        fp->close();
    }

    /* loading data to the buffer */
    void LoadDataToBuf();
};

/* 
 * A DataManager splits buffer to buckets and generates batches.
 * It must be associated with a dataset.
 */
class DataManager {
private:
    /* size of a mini-batch */
    int batchSize;

    /* the maximum length of a sequence */
    int maxSeqLen;

    /* the minimum length of a sequence */
    int minSeqLen;

    /* index of current bucket */
    int bucketIter = 0;

    /* the associated data set */
    DataSet* dataSet;

    /* buckets positon in buffered sequences */
    vector<vector<int>> buckets;

    DataManager() = delete;
    DataManager(DataManager&&) = delete;
    DataManager(const DataManager&) = delete;

    /* splits sequences to buckets */
    void BucketSeqByLen(int batchSize);

public:
    /* constructor */
    explicit DataManager(const char* fname,
        const char* srcVocab, const char* tgtVocab,
        int paraBufferSize, bool isShuffled,
        int paraMaxSeqLen, int paraMinSeqLen, int paraBatchSize)
    {
        batchSize = paraBatchSize;
        maxSeqLen = paraMaxSeqLen;
        minSeqLen = paraMinSeqLen;
        dataSet = new DataSet(fname, srcVocab, tgtVocab, paraBufferSize, isShuffled);
    }

    /* destructor */
    ~DataManager()
    {
        delete dataSet;
    }

    /* loads a mini-batch */
    void LoadBatch(XTensor* batchEnc,
        XTensor* paddingEnc, XTensor* batchDec,
        XTensor* paddingDec, XTensor* label,
        int devID, XMem* mem);
};

/* shuffle a file and pipe it to an output file */
void Shuffle(const char* srcFile);

} // namespace nts(NiuTrans.Tensor)

#endif // _DATA_READER_H_
