/* NiuTrans.Tensor - an open-source tensor library
 * Copyright (C) 2018, Natural Language Processing Lab, Northestern University.
 * All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/*
* $Created by: HU Chi (huchinlp@foxmail.com) 2019-03-18
*/

#include "T2TDataReader.h"
#include "StringUtil.h"

#include <algorithm>
#include <cstdio>
#include <fstream>



void nts::T2TDataSet::Process()
{
}


/* split buffer to buckets */
void nts::T2TDataManager::BucketSeqByLen(int batchSize)
{
    t2tDataSet->LoadDataToBuf();
    auto myCompare = [](const auto& a, const auto& b) {
        return a.size() < b.size();
    };
    sort(dataSet->dataBuf.begin(), dataSet->dataBuf.end(), myCompare);

    int realBatchSize = 0;
    for (int i = 0; i < dataSet->dataBuf.size(); ++i) {

        /* the first allocation */
        if (!bucketBoundary.size()) {
            bucketBoundary.emplace_back(1);
        }

        /* append the current sentence to bucket if the total size is suitable  */
        int seqLen = dataSet->dataBuf[i].first.size();
        int seqNum = bucketBoundary.back().size();
        if ((seqNum + 1) * seqLen < batchSize && seqLen <= maxSeqLen && seqLen >= minSeqLen) {
            bucketBoundary.back().push_back(i);
        }

        /* allocate a new bucket */
        else {
            bucketBoundary.emplace_back(1);
        }
    }

    /* drop the last bucket if its size is too small */
    if (bucketBoundary.back().size() * dataSet->dataBuf.back().first.size() < batchSize / 2)
        bucketBoundary.pop_back();
}

/*
load a batch of sequences (for MT)
>> batchEnc - the batch of the input sequences
>> paddingEnc - padding of the input sequences
>> batchDec - the batch of the output sequences
>> paddingDec - padding of the output sequences
>> devID - device id
>> mem - memory pool
>> isTraining - indicates whether we are training the model
*/
void nts::T2TDataManager::LoadBatchMT(XTensor* batchEnc, 
	XTensor* paddingEnc, XTensor* batchDec, 
	XTensor* paddingDec, XTensor* label, 
	int devID, XMem* mem, bool isTraining)
{

    if (bucketIter == bucketBoundary.size() - 1) {
        BucketSeqByLen(batchSize);
        bucketIter = 0;
    }
    auto bucket = bucketBoundary[bucketIter++];

    int sentenceCount = bucket.size();
    int maxEnc = dataSet->dataBuf[bucket.back()].first.size();
    int maxDec = dataSet->dataBuf[bucket.back()].second.size();

    InitTensor2D(batchEnc, sentenceCount, maxEnc, X_INT, devID, mem);
    InitTensor2D(paddingEnc, sentenceCount, maxEnc, X_FLOAT, devID, mem);
    InitTensor2D(batchDec, sentenceCount, maxDec, X_INT, devID, mem);
    InitTensor2D(paddingDec, sentenceCount, maxDec, X_FLOAT, devID, mem);
    InitTensor2D(label, sentenceCount, maxDec, X_INT, devID, mem);

    batchEnc->SetZeroAll();
    paddingEnc->SetZeroAll();
    batchDec->SetZeroAll();
    paddingDec->SetZeroAll();
    label->SetZeroAll();

    int* labelValues = new int[batchDec->unitNum];
    int* batchEncValues = new int[batchEnc->unitNum];
    int* batchDecValues = new int[batchDec->unitNum];
    MTYPE* paddingEncOffsets = new MTYPE[batchEnc->unitNum];
    MTYPE* paddingDecOffsets = new MTYPE[batchDec->unitNum];

    memset(batchEncValues, 0, sizeof(int) * batchEnc->unitNum);
    memset(batchDecValues, 0, sizeof(int) * batchDec->unitNum);
    memset(paddingEncOffsets, 0, sizeof(int) * batchEnc->unitNum);
    memset(paddingDecOffsets, 0, sizeof(int) * batchDec->unitNum);
    memset(labelValues, 0, sizeof(int) * batchDec->unitNum);

    int srcIter = 0;
    int tgtIter = 0;
    for (const auto& index : bucket) {
        int srcOffset = srcIter + dataSet->dataBuf[bucket.back()].first.size();
        int tgtOffset = tgtIter + dataSet->dataBuf[bucket.back()].second.size();
        for (const auto& src : dataSet->dataBuf[index].first)
            batchEncValues[srcIter++] = src;
        for (const auto& src : dataSet->dataBuf[index].second)
            batchDecValues[tgtIter++] = src;
        while (srcIter < srcOffset)
            paddingEncOffsets[srcIter++] = 1.0F;
        while (tgtIter < tgtOffset)
            paddingDecOffsets[tgtIter++] = 1.0F;
    }
    batchEnc->SetData(batchDecValues, batchEnc->unitNum);
    batchDec->SetData(batchDecValues, batchDec->unitNum);
    paddingEnc->SetData(paddingEncOffsets, batchEnc->unitNum);
    paddingDec->SetData(paddingDecOffsets, batchDec->unitNum);
    label->SetData(labelValues, label->unitNum);

    delete[] batchEncValues;
    delete[] batchDecValues;
    delete[] labelValues;
    delete[] paddingEncOffsets;
    delete[] paddingDecOffsets;
}
